@@ -29,8 +29,13 @@ through the code to commit the whitespace formatting.
 import copy
 
 from yapf.yapflib import format_token
+from yapf.yapflib import pytree_utils
+from yapf.yapflib import split_penalty
 from yapf.yapflib import style
 
+_COMPOUND_STMTS = frozenset({'for', 'while', 'if', 'elif', 'with', 'except',
+                             'def', 'class'})
+
 
 class FormatDecisionState(object):
   """The current state when indenting an unwrapped line.
@@ -65,6 +70,7 @@ class FormatDecisionState(object):
     """
     self.next_token = line.first
     self.column = first_indent
+    self.line = line
     self.paren_level = 0
     self.start_of_line_level = 0
     self.lowest_level_on_line = 0
@@ -88,9 +94,9 @@ class FormatDecisionState(object):
             self.column == other.column and
             self.paren_level == other.paren_level and
             self.start_of_line_level == other.start_of_line_level and
-            self.lowest_level_on_line == other.lowest_level_on_line and
-            (self.ignore_stack_for_comparison or
-             other.ignore_stack_for_comparison or self.stack == other.stack))
+            self.lowest_level_on_line == other.lowest_level_on_line and (
+                self.ignore_stack_for_comparison or
+                other.ignore_stack_for_comparison or self.stack == other.stack))
 
   def __ne__(self, other):
     return not self == other
@@ -108,52 +114,130 @@ class FormatDecisionState(object):
     """Returns True if the line can be split before the next token."""
     current = self.next_token
 
-    if not current.can_break_before:
+    if current.is_pseudo_paren:
       return False
 
-    return True
+    return current.can_break_before
 
   def MustSplit(self):
     """Returns True if the line must split before the next token."""
     current = self.next_token
-    previous_token = current.previous_token
+    previous = current.previous_token
+    column_limit = style.Get('COLUMN_LIMIT')
 
     if current.must_break_before:
       return True
 
+    if style.Get('DEDENT_CLOSING_BRACKETS'):
+      bracket = current if current.ClosesScope() else previous
+      if ((bracket.OpensScope() or bracket.ClosesScope()) and
+          format_token.Subtype.SUBSCRIPT_BRACKET not in bracket.subtypes):
+        if previous and previous.OpensScope():
+          length = (previous.matching_bracket.total_length -
+                    previous.total_length)
+          if _IsLastScopeInLine(previous.matching_bracket):
+            last_token = _LastTokenInLine(previous.matching_bracket)
+            length = last_token.total_length - previous.total_length
+
+          if length + previous.column >= column_limit:
+            return True
+        elif current.ClosesScope():
+          opening = current.matching_bracket
+          length = current.total_length - opening.total_length
+          if _IsLastScopeInLine(current):
+            last_token = _LastTokenInLine(current)
+            length = last_token.total_length - opening.total_length
+
+          if length + opening.column >= column_limit:
+            return True
+
     if (self.stack[-1].split_before_closing_bracket and
         # FIXME(morbo): Use the 'matching_bracket' instead of this.
         # FIXME(morbo): Don't forget about tuples!
         current.value in ']}'):
       # Split if we need to split before the closing bracket and the next
       # token is a closing bracket.
+      return current.node_split_penalty != split_penalty.UNBREAKABLE
+
+    if not previous:
+      return False
+
+    # TODO(morbo): This should be controlled with a knob.
+    if (format_token.Subtype.DICTIONARY_KEY in current.subtypes and
+        not current.is_comment):
+      # Place each dictionary entry on its own line.
       return True
 
-    if previous_token:
-      length = _GetLengthToMatchingParen(previous_token)
-      if (previous_token.value == '{' and  # TODO(morbo): List initializers?
-          length + self.column > style.Get('COLUMN_LIMIT')):
+    # TODO(morbo): This should be controlled with a knob.
+    if format_token.Subtype.DICT_SET_GENERATOR in current.subtypes:
+      return True
+
+    if (style.Get('SPLIT_BEFORE_NAMED_ASSIGNS') and
+        format_token.Subtype.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST in
+        current.subtypes):
+      if (previous.value not in {'(', '=', '*', '**'} and
+          current.value not in '=,)'):
+        opening = _GetOpeningParen(current)
+        if opening:
+          arglist_length = (
+              opening.matching_bracket.total_length - opening.total_length +
+              self.column)
+          return arglist_length > column_limit
+
+    if (previous.value in '{[' and current.lineno != previous.lineno and
+        format_token.Subtype.SUBSCRIPT_BRACKET not in previous.subtypes):
+      return True
+
+    if (previous.value == ':' and _IsDictionaryValue(current) and
+        current.lineno != previous.lineno):
+      # Retain the split between the dictionary key and value.
+      return True
+
+    if (format_token.Subtype.COMP_FOR in current.subtypes and
+        format_token.Subtype.COMP_FOR not in previous.subtypes):
+      length = _GetLengthOfSubtype(current, format_token.Subtype.COMP_FOR,
+                                   format_token.Subtype.COMP_IF)
+      if length + self.column > column_limit:
         return True
 
-      # TODO(morbo): This should be controlled with a knob.
-      if (format_token.Subtype.DICTIONARY_KEY in current.subtypes and
-          not current.is_comment):
-        # Place each dictionary entry on its own line.
+    if (format_token.Subtype.COMP_IF in current.subtypes and
+        format_token.Subtype.COMP_IF not in previous.subtypes):
+      length = _GetLengthOfSubtype(current, format_token.Subtype.COMP_IF)
+      if length + self.column > column_limit:
         return True
 
-      # TODO(morbo): This should be controlled with a knob.
-      if format_token.Subtype.DICT_SET_GENERATOR in current.subtypes:
-        return True
+    previous_previous_token = previous.previous_token
+    if (previous_previous_token and previous_previous_token.name == 'NAME' and
+        not previous_previous_token.is_keyword and previous.value == '('):
+      sibling = previous.node.next_sibling
+      if pytree_utils.NodeName(sibling) == 'arglist':
+        arglist = previous.node.next_sibling
+        if len(arglist.children) > 2:
+          if _IsFunctionCallWithArguments(current):
+            # There is a function call, with more than 1 argument, where
+            # the first argument is itself a function call with arguments.
+            # In this specific case, if we split after the first argument's
+            # opening '(', then the formatting will look bad for the rest
+            # of the arguments. Instead, enforce a split before that
+            # argument to keep things looking good.
+            return True
+          elif (current.OpensScope() and
+                current.matching_bracket.total_length + self.column >
+                column_limit):
+            # There is a data literal that will need to be split and could mess
+            # up the formatting.
+            return True
 
-      if (previous_token.value != '(' and
-          format_token.Subtype.DEFAULT_OR_NAMED_ASSIGN_ARG_LIST in
-          current.subtypes):
-        return style.Get('SPLIT_BEFORE_NAMED_ASSIGNS')
+    if (style.Get('SPLIT_BEFORE_BITWISE_OPERATOR') and
+        current.value in '&|' and previous.lineno < current.lineno):
+      return True
 
-      if (previous_token.value in '{[(' and
-          current.lineno != previous_token.lineno):
-        self.stack[-1].split_before_closing_bracket = True
-        return True
+    if (current.is_comment and
+        previous.lineno < current.lineno - current.value.count('\n')):
+      # If a comment comes in the middle of an unwrapped line (like an if
+      # conditional with comments interspersed), then we want to split if the
+      # original comments were on a separate line.
+      return True
 
     return False
 
@@ -203,14 +287,13 @@ class FormatDecisionState(object):
         #     foo = [a,
         #            b,
         #           ]
-        self.stack[-1].closing_scope_indent = previous.column
+        self.stack[-1].closing_scope_indent = self.column - 1
         if style.Get('ALIGN_CLOSING_BRACKET_WITH_VISUAL_INDENT'):
           self.stack[-1].closing_scope_indent += 1
         self.stack[-1].indent = self.column + spaces
       else:
         self.stack[-1].closing_scope_indent = (
-            self.stack[-1].indent - style.Get('CONTINUATION_INDENT_WIDTH')
-        )
+            self.stack[-1].indent - style.Get('CONTINUATION_INDENT_WIDTH'))
 
     self.column += spaces
 
@@ -260,11 +343,14 @@ class FormatDecisionState(object):
     # Calculate the split penalty.
     penalty = current.split_penalty
 
+    if previous.is_pseudo_paren and previous.value == '(':
+      # Small penalty for splitting after a pseudo paren.
+      penalty += 50
+
     # Add a penalty for each increasing newline we add.
     last = self.stack[-1]
-    penalty += (
-        style.Get('SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT') * last.num_line_splits
-    )
+    penalty += (style.Get('SPLIT_PENALTY_FOR_ADDED_LINE_SPLIT') *
+                last.num_line_splits)
     if not must_split and current.value not in {'if', 'for'}:
       # Don't penalize for a must split or for splitting before an
       # if-expression or list comprehension.
@@ -278,26 +364,35 @@ class FormatDecisionState(object):
     previous = current.previous_token
     top_of_stack = self.stack[-1]
 
-    if current.spaces_required_before > 2:
+    if current.spaces_required_before > 2 or self.line.disable:
       return current.spaces_required_before
 
     if current.OpensScope():
-      return self.first_indent if not self.paren_level else top_of_stack.indent
+      return top_of_stack.indent if self.paren_level else self.first_indent
 
     if current.ClosesScope():
       if (previous.OpensScope() or
           (previous.is_comment and previous.previous_token is not None and
            previous.previous_token.OpensScope())):
-        return max(
-            0, self.stack[-1].indent - style.Get('CONTINUATION_INDENT_WIDTH'))
+        return max(0,
+                   top_of_stack.indent - style.Get('CONTINUATION_INDENT_WIDTH'))
       return top_of_stack.closing_scope_indent
 
     if (previous and previous.is_string and current.is_string and
-        format_token.Subtype.DICTIONARY_VALUE in current.subtypes):
+        _IsDictionaryValue(current)):
       return previous.column
 
-    if format_token.Subtype.IF_TEST_EXPR in current.subtypes:
-      return top_of_stack.indent + style.Get('INDENT_IF_EXPR_CONTINUATION')
+    if style.Get('INDENT_DICTIONARY_VALUE'):
+      if _IsDictionaryValue(current):
+        if previous and (previous.value == ':' or previous.is_pseudo_paren):
+          return top_of_stack.indent
+
+    if (self.line.first.value in _COMPOUND_STMTS and
+        not style.Get('DEDENT_CLOSING_BRACKETS')):
+      token_indent = (len(self.line.first.whitespace_prefix.split('\n')[-1]) +
+                      style.Get('INDENT_WIDTH'))
+      if token_indent == top_of_stack.indent:
+        return top_of_stack.indent + style.Get('CONTINUATION_INDENT_WIDTH')
 
     return top_of_stack.indent
 
@@ -337,16 +432,16 @@ class FormatDecisionState(object):
     if is_multiline_string:
       # This is a multiline string. Only look at the first line.
       self.column += len(current.value.split('\n')[0])
-    else:
+    elif not current.is_pseudo_paren or current.value == '(':
       self.column += len(current.value)
 
     self.next_token = self.next_token.next_token
 
     # Calculate the penalty for overflowing the column limit.
     penalty = 0
-    if self.column > style.Get('COLUMN_LIMIT') and not current.is_comment:
+    if self.column > style.Get('COLUMN_LIMIT'):
       excess_characters = self.column - style.Get('COLUMN_LIMIT')
-      penalty = style.Get('SPLIT_PENALTY_EXCESS_CHARACTER') * excess_characters
+      penalty += style.Get('SPLIT_PENALTY_EXCESS_CHARACTER') * excess_characters
 
     if is_multiline_string:
       # If this is a multiline string, the column is actually the
@@ -356,22 +451,54 @@ class FormatDecisionState(object):
     return penalty
 
 
-def _GetLengthToMatchingParen(token):
-  """Returns the length from one bracket to the matching bracket.
+def _IsFunctionCallWithArguments(token):
+  while token:
+    if token.value == '(':
+      token = token.next_token
+      return token and token.value != ')'
+    elif token.name not in {'NAME', 'DOT'}:
+      break
+    token = token.next_token
+  return False
 
-  Arguments:
-    token: (FormatToken) The opening bracket token.
 
-  Returns:
-    The length to the closing paren or up to the first point where we can split
-    the line. The length includes the brackets.
-  """
-  if not token.matching_bracket:
-    return 0
-  end = token.matching_bracket
-  while end.next_token and not end.next_token.can_break_before:
-    end = end.next_token
-  return end.total_length - token.total_length + 1
+def _IsDictionaryValue(token):
+  while token:
+    if format_token.Subtype.DICTIONARY_VALUE in token.subtypes:
+      return True
+    token = token.previous_token
+  return False
+
+
+def _GetLengthOfSubtype(token, subtype, exclude=None):
+  current = token
+  while (current.next_token and subtype in current.subtypes and
+         (exclude is None or exclude not in current.subtypes)):
+    current = current.next_token
+  return current.total_length - token.total_length + 1
+
+
+def _GetOpeningParen(current):
+  previous = current
+  while previous is not None and previous.matching_bracket is None:
+    previous = previous.previous_token
+    if previous.ClosesScope():
+      previous = previous.matching_bracket.previous_token
+  return previous
+
+
+def _LastTokenInLine(current):
+  while current.next_token:
+    current = current.next_token
+  return current
+
+
+def _IsLastScopeInLine(current):
+  while current:
+    current = current.next_token
+    if current and current.OpensScope():
+      return False
+  return True
 
 
 class _ParenState(object):
