@@ -22,6 +22,7 @@ parser to perform the wrapping required to comply with the style guide.
 from lib2to3 import pytree
 
 from yapf.yapflib import format_token
+from yapf.yapflib import py3compat
 from yapf.yapflib import pytree_utils
 from yapf.yapflib import split_penalty
 from yapf.yapflib import style
@@ -73,6 +74,9 @@ class UnwrappedLine(object):
           _SpaceRequiredBetween(prev_token, token)):
         token.spaces_required_before = 1
 
+      token.total_length = (
+          prev_length + len(token.value) + token.spaces_required_before)
+
       # The split penalty has to be computed before {must|can}_break_before,
       # because these may use it for their decision.
       token.split_penalty += _SplitPenalty(prev_token, token)
@@ -80,10 +84,6 @@ class UnwrappedLine(object):
       token.can_break_before = (token.must_break_before or
                                 _CanBreakBefore(prev_token, token))
 
-      token.total_length = (
-          prev_length + len(token.value) + token.spaces_required_before
-      )
-
       prev_length = token.total_length
       prev_token = token
 
@@ -190,6 +190,13 @@ def _IsUnaryOperator(tok):
 
 def _SpaceRequiredBetween(left, right):
   """Return True if a space is required between the left and right token."""
+  if (left.is_pseudo_paren and _IsIdNumberStringToken(right) and
+      left.previous_token and _IsIdNumberStringToken(left.previous_token)):
+    # Space between keyword... tokens and pseudo parens.
+    return True
+  if left.is_pseudo_paren or right.is_pseudo_paren:
+    # The pseudo-parens shouldn't affect spacing.
+    return False
   if left.is_continuation or right.is_continuation:
     # The continuation node's value has all of the spaces it needs.
     return False
@@ -232,6 +239,9 @@ def _SpaceRequiredBetween(left, right):
     # No space between two unary operators.
     return False
   if left.is_binary_op or right.is_binary_op:
+    if left.value == '**' or right.value == '**':
+      # Space around the "power" operator.
+      return style.Get('SPACES_AROUND_POWER_OPERATOR')
     # Enforce spaces around binary operators.
     return True
   if (_IsUnaryOperator(left) and left.value != 'not' and
@@ -299,6 +309,10 @@ def _SpaceRequiredBetween(left, right):
   if right.value == ';':
     # Avoid spaces before a semicolon. (Why is there a semicolon?!)
     return False
+  if left.value == '(' and right.value == 'await':
+    # Special support for the 'await' keyword. Don't separate the 'await'
+    # keyword from an opening paren.
+    return False
   return True
 
 
@@ -320,6 +334,14 @@ def _MustBreakBefore(prev_token, cur_token):
 
 def _CanBreakBefore(prev_token, cur_token):
   """Return True if a line break may occur before the current token."""
+  if py3compat.PY3:
+    if prev_token.value == 'yield' and cur_token.value == 'from':
+      # Don't break before a yield argument.
+      return False
+    if (prev_token.value in {'async', 'await'} and
+        cur_token.value in {'def', 'with', 'for'}):
+      # Don't break after sync keywords.
+      return False
   if cur_token.split_penalty >= split_penalty.UNBREAKABLE:
     return False
   if prev_token.value == '@':
@@ -343,7 +365,9 @@ def _CanBreakBefore(prev_token, cur_token):
   if cur_token.is_comment and prev_token.lineno == cur_token.lineno:
     # Don't break a comment at the end of the line.
     return False
-  # TODO(morbo): There may be more to add here.
+  if format_token.Subtype.UNARY_OPERATOR in prev_token.subtypes:
+    # Don't break after a unary token.
+    return False
   return True
 
 
@@ -379,6 +403,7 @@ def _IsSurroundedByBrackets(tok):
 
 
 _LOGICAL_OPERATORS = frozenset({'and', 'or'})
+_BITWISE_OPERATORS = frozenset({'&', '|', '^'})
 _TERM_OPERATORS = frozenset({'*', '/', '%', '//'})
 
 
@@ -403,6 +428,19 @@ def _SplitPenalty(prev_token, cur_token):
     if cur_token.value in _LOGICAL_OPERATORS:
       return style.Get('SPLIT_PENALTY_LOGICAL_OPERATOR')
 
+  if style.Get('SPLIT_BEFORE_BITWISE_OPERATOR'):
+    # Prefer to split before '&', '|', and '^'.
+    if prev_token.value in _BITWISE_OPERATORS:
+      return style.Get('SPLIT_PENALTY_BITWISE_OPERATOR')
+    if cur_token.value in _BITWISE_OPERATORS:
+      return 0
+  else:
+    # Prefer to split after '&', '|', and '^'.
+    if prev_token.value in _BITWISE_OPERATORS:
+      return 0
+    if cur_token.value in _BITWISE_OPERATORS:
+      return style.Get('SPLIT_PENALTY_BITWISE_OPERATOR')
+
   if (format_token.Subtype.COMP_FOR in cur_token.subtypes or
       format_token.Subtype.COMP_IF in cur_token.subtypes):
     # We don't mind breaking before the 'for' or 'if' of a list comprehension.
@@ -413,7 +451,7 @@ def _SplitPenalty(prev_token, cur_token):
   if prev_token.value == ',':
     # Breaking after a comma is fine, if need be.
     return 0
-  if prev_token.value == '==':
+  if prev_token.is_binary_op:
     # We would rather not split after an equality operator.
     return 20
   if (format_token.Subtype.VARARGS_STAR in prev_token.subtypes or
@@ -436,6 +474,9 @@ def _SplitPenalty(prev_token, cur_token):
   if cur_token.value == '==':
     # We would rather not split before an equality operator.
     return split_penalty.STRONGLY_CONNECTED
+  if cur_token.ClosesScope():
+    # Give a slight penalty for splitting before the closing scope.
+    return 100
   if prev_token.value in _TERM_OPERATORS or cur_token.value in _TERM_OPERATORS:
     return 50
   return 0
